import typing as t


def str_join(
    *items: t.Any,
    sep: str = ", ",
    kv_sep: str = "=",
    str_func: t.Callable[[t.Any], str] = str,
    kvs: bool = False,
) -> str:
    if kvs:
        items = tuple(f"{k!s}{kv_sep}{str_func(v)}" for k, v in items)
        str_func = str
    return sep.join(str_func(item) for item in items)


def clsqualname(obj: object) -> str:
    if isinstance(obj, type):
        return obj.__qualname__
    return obj.__class__.__qualname__


def to_repr(
    obj: object,
    *,
    args: str | t.Iterable | None = None,
    kwargs: str | t.Mapping | None = None,
) -> str:
    params = []
    if args:
        if not isinstance(args, str):
            args = str_join(*args, str_func=repr)
        params.append(args)
    if kwargs:
        if not isinstance(kwargs, str):
            kwargs = str_join(*kwargs.items(), str_func=repr, kvs=True)
        params.append(kwargs)
    return f"{clsqualname(obj)}({str_join(*params)})"


def camel_to_snake(string: str) -> str:
    out = []
    for i, ch in enumerate(string):
        if ch.isupper() and i:
            out.append("_")
        out.append(ch.lower())
    return "".join(out)



import contextlib
import typing as t


# class Community:
#     def __init__(self, value: int):  # ???
#         ...
#
#     @classmethod
#     def from_string(cls, val: str):
#         ...


class ASN:

    """BGP ASN data type

    https://networklessons.com/bgp/bgp-4-byte-number
    """

    _SEP = "."
    _ZERO = 0
    _TWO_BYTES = 2 ** 16
    _FOUR_BYTES = 2 ** 32

    def __init__(self, number: int):
        if not isinstance(number, int):
            raise TypeError(f"int expected, got: {type(number)}")

        if not (self._ZERO <= number < self._FOUR_BYTES):
            msg = f"Value must be in range {self._ZERO}-{self._FOUR_BYTES - 1}"
            raise ValueError(msg)

        self._number = number

    @classmethod
    def from_plain(cls, string: str) -> t.Self:
        if not isinstance(string, str):
            raise TypeError(f"str expected, got: {type(string)}")

        return cls(int(string))

    def to_plain(self) -> str:
        return str(self._number)

    @classmethod
    def from_asdot(cls, string: str) -> t.Self:
        with contextlib.suppress(Exception):
            result = cls.from_plain(string)
            if int(result) < cls._TWO_BYTES:
                return result
        return cls.from_asdotplus(string)

    def to_asdot(self) -> str:
        if self._number < self._TWO_BYTES:
            return self.to_plain()
        return self.to_asdotplus()

    @classmethod
    def from_asdotplus(cls, string: str) -> t.Self:
        if not isinstance(string, str):
            raise TypeError(f"str expected, got: {type(string)}")

        high, low = map(int, string.split(cls._SEP, maxsplit=1))
        return cls(high * cls._TWO_BYTES + low)

    def to_asdotplus(self) -> str:
        return self._SEP.join(map(str, divmod(self._number, self._TWO_BYTES)))

    @classmethod
    def parse(cls, data: int | str) -> t.Self:
        with contextlib.suppress(Exception):
            return cls(t.cast(int, data))
        data = t.cast(str, data)
        with contextlib.suppress(Exception):
            return cls.from_plain(data)
        with contextlib.suppress(Exception):
            return cls.from_asdotplus(data)
        raise ValueError(f"Unable to parse data into {cls}")

    def __int__(self) -> int:
        return self._number

    def __str__(self) -> str:
        return f"<{self.__class__.__name__}: {self.to_asdotplus()}>"

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self._number})"

    def __copy__(self):
        return type(self)(self._number)

    def __deepcopy__(self, memo: dict[int, t.Any]):
        _id = id(self)
        if _id not in memo:
            memo[_id] = type(self)(self._number)
        return memo[_id]

    def __reduce__(self) -> tuple[t.Any, ...]:
        return self.__class__, (self._number,)

    def __hash__(self) -> int:
        return hash(self._number)

    def __eq__(self, other) -> bool:
        return (isinstance(other, self.__class__)
                and (int(other) == self._number))


# print(ASN(6541))   # 0.6541
# print(ASN(54233))  # 0.54233
# print(ASN(544))    # 0.544
#
# print(ASN(65536))    # 1.0
# print(ASN(65537))    # 1.1
# print(ASN(65538))    # 1.2



import types
import typing as t


class BaseRegistryError(Exception):
    pass


class SubclassMissingQualifier(BaseRegistryError):
    pass


class SubclassNotFound(BaseRegistryError):
    pass


class ClassRegistriesMixin:
    __registries: dict[str, dict[t.Any, type]] = {}

    __registry_type__: str  # default: class name
    __registry_qualifier__: str | t.Callable[[t.Type], str]

    @classmethod
    def __get_type(cls):
        return getattr(cls, "__registry_type__", cls.__name__)

    @classmethod
    def get_qualifier(cls):
        qualifier = getattr(cls, "__registry_qualifier__", None)
        if qualifier is None:
            raise SubclassMissingQualifier(cls)
        return qualifier

    @classmethod
    def get_identity(cls):
        qualifier = cls.get_qualifier()
        if isinstance(qualifier, str):
            identity = getattr(cls, qualifier, None)
        else:
            identity = qualifier(cls)

        return identity

    def __init_subclass__(cls, **kwargs: t.Any):
        super().__init_subclass__(**kwargs)
        mixin = ClassRegistriesMixin

        if mixin in cls.__bases__:
            cls.__registry_type__ = cls.__get_type()  # enrich
            cls.get_qualifier()  # validation
        else:
            store = mixin.__registries.setdefault(cls.__get_type(), {})
            store[cls.get_identity()] = cls

    @classmethod
    def get_subclass_for(cls, identity: t.Any) -> type:
        store = cls.__registries[cls.__registry_type__]
        if identity not in store:
            raise SubclassNotFound(identity)
        return store[identity]

    @classmethod
    def list_subclasses(cls) -> types.MappingProxyType[str, type]:
        store = cls.__registries[cls.__registry_type__]
        return types.MappingProxyType(store.copy())



from marshmallow import validate


class Lambda(validate.Validator):
    def __init__(self, func):
        self.func = func

    def __call__(self, value):
        if self.func(value):
            return value

        raise validate.ValidationError("Lambda failed")


class HasKeys(validate.Validator):
    def __init__(self, *keys):
        self.keys = set(keys)

    def __call__(self, value):
        if self.keys.issubset(value):
            return value

        msg = f"Missing keys: {', '.join(self.keys)}"
        raise validate.ValidationError(msg)



import dataclasses
import logging


# NOTE(d.burmistrov): this module is DEPRECATED
logging.getLogger(__name__).warning("Module `dataclasses` is deprecated in Miska")  # noqa


_K_PRIVATE = "private"
_K_METADATA = "metadata"
_BASE_MODEL_INIT_FLAG = "_BASE_MODEL_INIT_FLAG"


def public_field(**kwargs) -> dataclasses.Field:
    kwargs.setdefault(_K_METADATA, {})[_K_PRIVATE] = False
    return dataclasses.field(**kwargs)


def private_field(**kwargs) -> dataclasses.Field:
    kwargs.setdefault(_K_METADATA, {})[_K_PRIVATE] = True
    return dataclasses.field(**kwargs)


class BaseModel:

    """Base class for dataclass magic

    - supports private fields that can't be modified

    from dataclasses import *
    from miska.dataclasses import *

    @dataclass
    class MyDataClass(BaseModel):
        public_field: str = public_field()
        private_field: str = private_field()

    d = MyDataClass("a", "b")
    print(d)  # MyDataClass(public_field='a', private_field='b')

    d.public_field = "A"
    print(d)  # MyDataClass(public_field='A', private_field='b')

    d.private_field = "B"
    # AttributeError: Forbidden to modify private attribute: private_field
    """

    __allow_attribute_deletion__ = False
    __default_private__ = True

    __PRIVATES = {}

    as_tuple = dataclasses.astuple
    as_dict = dataclasses.asdict

    def __post_init__(self):
        self._validate()
        self._mark_init_finished()

    def _mark_init_finished(self) -> None:
        setattr(self, _BASE_MODEL_INIT_FLAG, False)

    def _setattr(self, key, value):
        return super().__setattr__(key, value)

    def _validate(self) -> None:
        pass

    def __discover_privates(self):
        privates = set()
        for field in dataclasses.fields(self):
            flag = field.metadata.get(_K_PRIVATE, self.__default_private__)
            if flag:
                privates.add(field.name)
        return privates

    def __get_privates(self):
        privates = self.__PRIVATES.get(self.__class__)
        if privates is None:
            privates = self.__discover_privates()
            self.__PRIVATES[self.__class__] = privates
        return privates

    def __setattr__(self, key: str, value):
        # allow to initialize instance during dataclass magic
        if getattr(self, _BASE_MODEL_INIT_FLAG, True):
            return super().__setattr__(key, value)
        # allow managing regular private attributes
        elif key.startswith("_") and hasattr(self, key):
            return super().__setattr__(key, value)

        privates = self.__get_privates()

        # magic for inner modification like `self._private_field = 42`
        #   (it converts name into field that marked as private)
        if key.startswith("_"):
            _key = key[1:]
            if _key in privates:
                return super().__setattr__(_key, value)

        # magic for restricting private field modifications
        if key in privates:
            msg = f"Forbidden to modify private attribute: {key}"
            raise AttributeError(msg)

        return super().__setattr__(key, value)

    def __delattr__(self, item):
        if self.__allow_attribute_deletion__:
            return super().__delattr__(item)

        raise AttributeError("Attributes can't be deleted")



import collections
import itertools
import operator
import typing as t


class GroupByAttr:
    def __init__(self,
                 attr_name: str,
                 grouping_map: dict[str, t.Any],
                 default_group: str = "others"):
        self._attr_name = attr_name
        self._default_group = default_group
        self._grouping = dict(map(reversed, grouping_map.items()))
        self._key_func = operator.attrgetter(self._attr_name)

    def __call__(self, iterable):
        iterable = sorted(iterable, key=self._key_func)
        grouped = itertools.groupby(iterable, key=self._key_func)

        result = collections.defaultdict(list)
        for key_value, it in grouped:
            key = self._grouping.get(key_value, self._default_group)
            result[key].extend(it)

        return result


def interseq(ileft: t.Iterable,
             iright: t.Iterable,
             /, *,
             key: t.Callable[[t.Any], t.Hashable] = lambda item: item):
    left_map = {key(item): item for item in ileft}
    right_map = {key(item): item for item in iright}
    common_keys = set(left_map).intersection(right_map)
    left = tuple(v for k, v in left_map.items() if k not in common_keys)
    right = tuple(v for k, v in right_map.items() if k not in common_keys)
    common = tuple((left_map[key], right_map[key]) for key in common_keys)
    return left, common, right



class Comparable:
    def __init__(self, obj, comp):
        self.obj = obj
        self.value = self._make_value(comp)
        self._hash = hash(self.value)

    def _make_value(self, comp):
        return comp(self.obj) if callable(comp) else comp

    def __hash__(self):
        return self._hash

    def __eq__(self, other):
        if not isinstance(other, self.__class__):
            msg = f"Can't compare {self.__class__} with {other.__class__}"
            raise TypeError(msg)

        return self.value == other.value



import enum


class PreserveCaseStrEnum(enum.StrEnum):
    """Copy enum names into their values

    class StatusEnum(PreserveCaseStrEnum):
         ok = auto()       # StatusEnum.ok.value == "ok"
         FAILED = auto( )  # StatusEnum.FAILED.value == "FAILED"
         Unknown = auto()  # StatusEnum.Unknown.value == "Unknown"
    """

    @staticmethod
    def _generate_next_value_(name, start, count, last_values) -> str:
        return name


def member_on_missing(member_name):
    """Provide default member on missing

    class StatusEnum(str, enum.Enum):
         OK = 'ok'
         FAILED = 'failed'
         _missing_ = enums.member_on_missing("FAILED")

    StatusEnum.OK  # <StatusEnum.OK: 'ok'>
    StatusEnum.FAILED  # <StatusEnum.FAILED: 'failed'>

    StatusEnum("ok")  # <StatusEnum.OK: 'ok'>
    StatusEnum("failed")  # <StatusEnum.FAILED: 'failed'>

    # HOW IT WORKS
    StatusEnum("unknown")  # <StatusEnum.FAILED: 'failed'>
    """

    return classmethod(lambda cls, value: getattr(cls, member_name))



import types

import marshmallow as mm
from marshmallow import exceptions as excs
from marshmallow import utils


def _wrap(wrap_type):
    def decorator(func):
        return lambda *args, **kwargs: wrap_type(func(*args, **kwargs))

    return decorator


class MappingProxy(mm.fields.Dict):
    _serialize = _wrap(dict)(mm.fields.Dict._serialize)
    _deserialize = _wrap(types.MappingProxyType)(mm.fields.Dict._deserialize)


class TensileTuple(mm.fields.Field):
    #: Default error messages.
    default_error_messages = {"invalid": "Not a valid tuple."}

    def __init__(self, item_type, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if not isinstance(item_type, mm.fields.Field):
            msg = ("Value of 'item_type' must be subclasses"
                   " or instances of marshmallow.base.FieldABC.")
            raise ValueError(msg) from exc
        self.item_type = item_type

    def _serialize(self, value, attr, obj, **kwargs) -> tuple | None:
        if value is None:
            return None

        return tuple(self.item_type._serialize(item, attr, obj, **kwargs)
                     for item in value)

    def _deserialize(self, value, attr, data, **kwargs) -> tuple:
        if not mm.utils.is_collection(value):
            raise self.make_error("invalid")

        result = []
        errors = {}
        for idx, item in enumerate(value):
            try:
                result.append(self.item_type.deserialize(item, **kwargs))
            except mm.ValidationError as error:
                if error.valid_data is None:
                    errors.update(idx=error.messages)
                else:
                    result.append(error.valid_data)
        if errors:
            raise mm.ValidationError(errors, valid_data=result)

        return tuple(result)



import typing as t


def construct_filtered_dict(
    func: t.Callable[[t.Any], bool] = bool,
    /,
    **kwargs: t.Any,
) -> dict[str, t.Any]:
    return {k: v for k, v in kwargs.items() if func(v)}


def wipe(
    data: dict,
    *,
    keys: list[t.Hashable] | tuple[t.Hashable, ...] | t.Hashable,
    copy: bool = False,
) -> dict:
    if copy:
        data = data.copy()

    if not isinstance(keys, (tuple, list)):
        keys = (keys,)
    for k in keys:
        data.pop(k, None)

    return data


def rekey(
    data: dict,
    *,
    key: t.Hashable,
    new_key: t.Hashable,
    copy: bool = False,
) -> dict:
    if copy:
        data = data.copy()

    value = data.pop(key)
    data[new_key] = value

    return data


def new(data: dict, *exclude, **updates) -> dict:
    new_data = data.copy()
    for ex in exclude:
        new_data.pop(ex, None)
    new_data.update(updates)
    return new_data


def mutate_values(
    data: dict,
    *,
    keys: t.Sequence[t.Hashable],
    func: t.Callable[[t.Any], t.Any],
    copy: bool = False,
) -> dict:
    if copy:
        data = data.copy()
    for k in keys:
        data[k] = func(data[k])
    return data



class KwargsOnly:

    def __new__(cls, *args, **kwargs):
        if args:
            raise TypeError("%s only accepts kwargs" % cls)
        inst = super().__new__(cls)
        inst.__kwargs = kwargs
        return inst
